# -*- coding: utf-8 -*-
import logging
import time
import uuid

import pika
from .base import Connector
import pickle
import json

from .exceptions import (ERROR_FLAG, HAS_ERROR, NO_ERROR, RemoteFunctionError,
                         RemoteCallTimeout)

logger = logging.getLogger(__name__)

class RPCClient(Connector):
    '''
    RPC client.
    Parameters:
        amqp_url: amqp url connect to rabbitmq server. If this is specified, host ip and port parameters will not be used.

        host, port, username, passwd: Rabbitmq ip address and port and account.
        durable, auto_delete: The exchange property that this client will connect to. But, RPCClient will create its own
            queue, whose name is random allocated. This random allocated queue will not be affected by these parameters.
        virtual_host: Your vhost in rabbitmq server
        threaded: Invalid in RPCClient
        exchange: exchange name that will be used.
        queue_name: queue name you want to connect. If not setted, the queue name will be allocated randomly.
        bDataJson: Whether your data will be transmitted in json. If setted to false, the data will be transmitted with pickle.
            This flag should be set if the RPCServer also set to True, otherwise, leave it default.
    '''
    def __init__(self, bDataJson = False, queue_name = "",**kwargs):
        self._results = {}
        self.callback_queue = None
        self.bDataJson = bDataJson
        self._local_queues = []
        self.queue_name = queue_name
        super(RPCClient, self).__init__(**kwargs)
        self._threaded = False # Force threaded flag to false
        self._connection = self.connect()
        self._channel = self._connection.channel()
        self._channel.exchange_declare(self._exchange, exchange_type='direct', auto_delete=self.auto_delete, durable=self.durable)
        self.setup_callback_queue()

    def setup_callback_queue(self):
        if not self.callback_queue:
            if len(self.queue_name):
                ret = self._channel.queue_declare(queue=self.queue_name, exclusive=False, auto_delete=self.auto_delete,
                                                  durable=self.durable)
            else:
                ret = self._channel.queue_declare(queue=self.queue_name, exclusive=False, auto_delete=True)
            self.callback_queue = ret.method.queue
            self._channel.queue_bind(self.callback_queue, self._exchange)
            self._channel.basic_consume(self.callback_queue,
                                       self.on_response,
                                       auto_ack=True)

    def on_response(self, channel, basic_deliver, props, body):
        ret = pickle.loads(body)
        if props.headers.get(ERROR_FLAG, NO_ERROR) == HAS_ERROR:
            ret = RemoteFunctionError(ret)
        self._results[props.correlation_id] = ret

    def get_response(self, correlation_id, timeout=None):
        stoploop = time.time() + timeout if timeout is not None else 0
        while stoploop > time.time() or timeout is None:
            self._connection.process_data_events()
            if correlation_id in self._results:
                return self._results.pop(correlation_id)
            self._connection.sleep(0.01)

        raise RemoteCallTimeout()


    def publish_message(self, exchange, routing_key, body, ignore_result = False, headers=None):
        corr_id = str(uuid.uuid4())
        rply_to = None
        if not ignore_result:
            rply_to = self.callback_queue
        if self.bDataJson:
            body = json.dumps(body).encode('utf-8')
        else:
            body = pickle.dumps(body)

        self._channel.basic_publish(
            exchange=exchange,
            routing_key=routing_key,
            properties=pika.BasicProperties(
                reply_to=rply_to,
                headers=headers,
                correlation_id=corr_id,
            ),
            body=body)

        return corr_id

    def skip_response(self, correlation_id):
        self._results.pop(correlation_id, None)

    def call(self, consumer_name):

        def func(*args, **kwargs):
            """Call the remote function.

            :param bool ignore_result: Ignore the result return immediately.
            :param str exchange: The exchange name consists of a non-empty.
            :param str routing_key: The routing key to bind on.
            :param float timeout: if waiting the result over timeount seconds,
                                  RemoteCallTimeout will be raised .
            """
            ignore_result = kwargs.pop('ignore_result', False)
            exchange = kwargs.pop('exchange', self._exchange)
            routing_key = kwargs.pop('routing_key', self._exchange)
            timeout = kwargs.pop('timeout', None)

            if not ignore_result:
                self.setup_callback_queue()

            try:
                if timeout is not None:
                    timeout = float(timeout)
            except (ValueError, TypeError):
                raise ValueError("'timeout' is expected a float.")

            payload = {
                'args': args,
                'kwargs': kwargs,
            }
            corr_id = self.publish_message(
                exchange,
                routing_key,
                body=payload,
                ignore_result = ignore_result,
                headers={'consumer_name': consumer_name})

            logger.info('Sent remote call: %s', consumer_name)
            if not ignore_result:
                try:
                    ret = self.get_response(corr_id, timeout)
                except RemoteCallTimeout:
                    raise RemoteCallTimeout(
                        "Calling remote function '%s' timeout." % consumer_name)

                if isinstance(ret, RemoteFunctionError):
                    raise ret

                return ret

            self.skip_response(corr_id)

        func.__name__ = consumer_name
        return func

    def __getattribute__(self, key):
        if key.startswith('call_'):
            _, consumer_name = key.split('call_')
            return self.call(consumer_name)

        return super(RPCClient, self).__getattribute__(key)

    def __del__(self):
        pass
        # if not self._connection.:
        #     self._connection.close()



# -*- coding: utf-8 -*-
import logging
import sys

from .client import RPCClient
from .server import RPCServer
from .consumer import consumer, Consumer
from .credentials import AliyunCredentialsProvider

LOG_FORMAT = (
    '%(levelname)s %(asctime)s %(name)s %(funcName)s %(lineno)s: %(message)s')
logging.basicConfig(level=logging.ERROR, format=LOG_FORMAT)

__all__ = [
    'consumer', 'Consumer', 'RPCClient', 'RPCServer',
    'AliyunCredentialsProvider'
]


def main():
    from rabbit_rpc.commands import ManageUtility
    manage = ManageUtility(sys.argv)
    manage.execute()

if __name__ == '__main__':
    main()



import time
import threading
import queue
import functools

class LockNames():
    names = {}

class ThreadAtomLock(object):
    '''
    适用于需要原子操作的函数。直接修饰该函数即可
    用法：
    @ThreadAtomLock()
    def func(a,b,c):
        ...
    则该func在多线程的调用中会被自动加锁
    '''
    def __init__(self, lockname = "default"):
        self.lockname = lockname
        if lockname in LockNames.names:
            self.lock = LockNames.names[lockname]
        else:
            LockNames.names[lockname] = threading.Lock()
            self.lock = LockNames.names[lockname]

    def __call__(self, func):
        def _atom(*arg, **kw):
            self.lock.acquire()
            try:
                v = func(*arg, **kw)
            except Exception as e:
                self.lock.release()
                raise
            self.lock.release()
            return v
        return _atom
    
    def __enter__(self):
        self.lock.acquire()
    
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.lock.release()

class ThreadPool(object):
    '''
    定义线程池类。
    参数：
        pool_size为线程池中能够容纳的线程个数。
        bReturnObj:修饰的函数是否返回ThreadPool对象，返回的对象可以用于调用join，等待所有线程结束
    使用方法：
        @ThreadPool(3, bReturnObj = True)
        def func(a,b,c):
            ...
        for i in range(n):
            th_obj = func(a,b,c)
        th_obj.join() # 等待所有线程结束
        在调用func时，func会使用多线程执行，但一次最多开启3个线程。当线程池中有线程结束后，
        才允许新的线程加入线程池
    '''
    def __init__(self, pool_size, bReturnObj = False):
        self.pool_size = pool_size
        self.flag_pool = queue.Queue(pool_size)
        self._thread_in_pool = [None]*pool_size
        self.lock = threading.Lock()
        self.bRetObj = bReturnObj
    def _allocate_empty_pos(self):
        for i,th in enumerate(self._thread_in_pool):
            if th is None:
                return i
        return None


    def __call__(self, func):
        def _run_thread_func(th_pos, func, *args, **kwargs):
            func(*args, **kwargs)
            self._thread_in_pool[th_pos] = None
            self.flag_pool.get()
        @functools.wraps(func)
        def thread_func(*args, **kwargs):
            self.flag_pool.put('V')
            th_pos = 0
            th_pos = self._allocate_empty_pos()
            if th_pos is None:
                raise Exception("Error pos")
            th = threading.Thread(target=_run_thread_func, args=(th_pos, func, *args), kwargs=kwargs)
            self._thread_in_pool[th_pos] = th
            th.start()
            if self.bRetObj:
                return self
        return thread_func

    def join(self):
        while any(self._thread_in_pool):
            time.sleep(0.05)
__all__ = ["ThreadAtomLock", "ThreadPool"]



# -*- coding: utf-8 -*
import hmac
import base64
import time
import hashlib


class AliyunCredentialsProvider:
    """
    根据阿里云的 accessKey, accessSecret, UID算出 amqp 连接使用的 username 和
    password。 UID是资源ownerID，一般是接入点第一段
    """
    ACCESS_FROM_USER = 0

    def __init__(self, access_key, access_secret, uid):
        self.accessKey = access_key
        self.accessSecret = access_secret
        self.UID = uid

    def get_username(self):
        t = '%i:%s:%s' % (self.ACCESS_FROM_USER, self.UID, self.accessKey)
        return base64.b64encode(t.encode('utf-8'))

    def get_password(self):
        ts = str(int(round(time.time() * 1000)))
        h = hmac.new(
            ts.encode('utf-8'), self.accessSecret.encode('utf-8'), hashlib.sha1)
        sig = h.hexdigest().upper()
        sig_str = "%s:%s" % (sig, ts)
        return base64.b64encode(sig_str.encode('utf-8'))



# -*- coding: utf-8 -*-


class Queue(object):

    def __init__(self, name, dispatcher, exclusive=False):
        self.name = name
        self.exclusive = exclusive
        self.dispatcher = dispatcher

    def add_consumer(self, consumer):
        return self.dispatcher.register(consumer)



# -*- coding: utf-8 -*-
import logging
import pickle
import json

import pika
from concurrent.futures import ThreadPoolExecutor
from six import python_2_unicode_compatible

from .threadtool import ThreadAtomLock

from .exceptions import ERROR_FLAG, HAS_ERROR, NO_ERROR

logger = logging.getLogger(__name__)


@python_2_unicode_compatible
class Consumer(object):

    def __init__(self, name, queue=None, exclusive=False):
        self.name = name
        self.queue = queue
        self.exclusive = exclusive
        self.bJsonParameters = False

    def consume(self, *args, **kwargs):
        pass

    def __str__(self):
        return self.name

    def __repr__(self):
        return '<%s.Consumer: %s>' % (self.__module__, self.name)


def consumer(name=None, queue=None, exclusive=False):

    def decorator(func):
        cname = name or func.__name__

        c = Consumer(cname, queue, exclusive)
        c.consume = func
        return c

    return decorator

ReplyLockName = "_MsgReply"
class MessageDispatcher(object):
    def __init__(self, channel, exchange='', threaded = True, threadpool_size = -1):
        self._channel = channel
        self.threads = []
        self._registries = {}
        if threadpool_size <= 0:
            self._executor = ThreadPoolExecutor()
        else:
            self._executor = ThreadPoolExecutor(threadpool_size)
        self._exchange = exchange
        self._threaded = threaded

        self.consumer_tag = None

    def register(self, consumer):
        if consumer.name not in self._registries:
            self._registries[consumer.name] = consumer

    def __call__(self, *args, **kwargs):
        return self.dispatch_message(*args, **kwargs)

    def clear(self):
        self._registries = {}

    def dispatch_message(self, channel, basic_deliver, properties, body):
        """Invoked by pika when a message is delivered from RabbitMQ. The
        channel is passed for your convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param pika.channel.Channel channel: The channel object
        :param pika.Spec.Basic.Deliver: basic_deliver method
        :param pika.Spec.BasicProperties: properties
        :param Bytes or json: The message body

        """
        try:
            consumer_name = properties.headers.get('consumer_name')
        except :
            consumer_name = 'default'
        if consumer_name is None:
            consumer_name = 'default'

        logger.info("Received a remote call on function '%s'", consumer_name)

        try:
            consumer = self._registries[consumer_name]
        except KeyError:
            msg = "Function '%s' not found." % consumer_name
            logger.info(msg)
            if properties.reply_to:
                self.reply_message(
                    properties, msg, is_error=True)

            self.acknowledge_message(basic_deliver.delivery_tag)
            return
        if not consumer.bJsonParameters:
            arguments = pickle.loads(body)
        else:
            try:
                arguments = json.loads(body.decode("utf-8"))
            except Exception as e:
                logger.error("Load arguments failed: {}".format(e))
                arguments = {}
        args = arguments.get('args', [])
        kwargs = arguments.get('kwargs', {})
        if len(args) == 0 and len(kwargs) == 0 and len(arguments):
            kwargs = arguments
        if not self._threaded:
            self.call_comsumer(consumer, basic_deliver.delivery_tag, properties, *args, **kwargs)
        else:
            self._executor.submit(self.call_comsumer, consumer,
                                basic_deliver.delivery_tag, properties, *args, **kwargs)
    @ThreadAtomLock(ReplyLockName)
    def reply_message(self, props, body, headers=None, is_error=False):
        if headers is None:
            headers = {}

        headers[ERROR_FLAG] = NO_ERROR if not is_error else HAS_ERROR

        self._channel.basic_publish(
            exchange=self._exchange,
            routing_key=props.reply_to,
            properties=pika.BasicProperties(
                correlation_id=props.correlation_id,
                headers=headers),
            body=pickle.dumps(body))

    def call_comsumer(self, consumer, delivery_tag, props, *args, **kwargs):
        try:
            ret = consumer.consume(*args, **kwargs)
            is_error = False
        except Exception as ex:
            logger.exception(
                'Error occurred when calling consumer. consumer: %s, args: %s, '
                'kwargs: %s', consumer.name, args, kwargs)
            ret = str(ex)
            is_error = True

        if props.reply_to is not None:
            self.reply_message(props, ret, is_error=is_error)

        self.acknowledge_message(delivery_tag)

    @ThreadAtomLock(ReplyLockName)
    def acknowledge_message(self, delivery_tag):
        ret = self._channel.basic_ack(delivery_tag)

    def __contains__(self, consumer_name):
        return consumer_name in self._registries

    def stop(self):
        self._executor.shutdown()



import six

if six.PY3:
    from threading import Condition

else:
    from threading import _Condition
    from monotonic import monotonic as _time

    class Condition(_Condition):

        def wait_for(self, predicate, timeout=None):
            """Wait until a condition evaluates to True.

            predicate should be a callable which result will be interpreted as a
            boolean value.  A timeout may be provided giving the maximum time to
            wait.

            """
            endtime = None
            waittime = timeout
            result = predicate()
            while not result:
                print(result)
                if waittime is not None:
                    if endtime is None:
                        endtime = _time() + waittime
                    else:
                        waittime = endtime - _time()
                        if waittime <= 0:
                            break
                self.wait(waittime)
                result = predicate()
            return result


# -*- coding: utf-8 -*-
from argparse import ArgumentParser
import sys


def get_commands():
    from .worker import Worker
    return {Worker.name: Worker()}


class ManageUtility(object):

    def __init__(self, argv=None):
        self.argv = argv

    def fetch_command(self, subcommand):
        commands = get_commands()
        try:
            return commands[subcommand]
        except KeyError:
            pass

    def execute(self):
        parser = ArgumentParser()
        parser.add_argument('worker', help='start a server worker')
        # parser.add_argument('call', help='send remote call')

        try:
            subcommand = self.argv[1]
            command = self.fetch_command(subcommand)
            if not command:
                parser.print_help()
                sys.stderr.write('Unkown subcommand: %s\n' % subcommand)
                sys.exit(1)
        except IndexError:
            parser.print_help()
            sys.exit()

        try:
            command = self.fetch_command(subcommand)
            command.run_from_argv(*self.argv[2:])
        except KeyboardInterrupt:
            pass



# -*- coding: utf-8 -*-
import logging

from .base import Connector
from .consumer import MessageDispatcher,Consumer
from .queue import Queue

logger = logging.getLogger(__name__)

class RPCServer(Connector):
    '''
    RPC Server class
    Parameters:
    queue_name: queue name in rabbitmq server
    consumers: optional parameters. You can pass list of Consumer objects.
    threaded: True or False. If setted True, means run in multi-threads mode, otherwise in blocking mode.
    amqp_url: Please refer to base class 'Connector'
    host,port,username, passwd, exchange: Please refer to base class 'Connector'
    num_threads: If threaded==True, RPC server can run in multi-threads mode. Then you can specify max threads you want.
        Default -1, means automatically decide number of threads.
    '''

    def __init__(self,queue_name = None, consumers = None, num_threads=-1, durable = False, auto_delete = True, *args, **kwargs):
        self._queues = {}
        if consumers is None:
            self._consumers = []
        else:
            self._consumers = consumers
        self.default_queue = queue_name or self.DEFUALT_QUEUE
        self.num_threads =num_threads
        if num_threads > 0:
            prefetch_count = num_threads
        else:
            prefetch_count = 1
        super(RPCServer, self).__init__(durable=durable, auto_delete=auto_delete,prefetch_count=prefetch_count,
                                        *args, **kwargs)
    def consumer(self, name=None, queue=None, exclusive=False, bJsonArgs = False):
        def decorator(func):
            cname = name or func.__name__
            c = Consumer(cname, queue, exclusive)
            c.consume = func
            c.bJsonParameters = bJsonArgs
            self._consumers.append(c)
            return func
        return decorator

    def on_exchange_declareok(self, unused_frame):
        self.setup_queues()

    def _setup_queue(self, queue_name):
        dispatcher = MessageDispatcher(self._channel, self._exchange, threaded=self._threaded,
                                       threadpool_size=self.num_threads)
        queue = Queue(queue_name, dispatcher)
        self._queues[queue_name] = queue
        return queue

    def setup_default_queue(self):
        return self._setup_queue(self.default_queue)

    def setup_queues(self):
        """Setup the queue on RabbitMQ by invoking the Queue.Declare RPC
        command.

        :param str|unicode queue_name: The name of the queue to declare.
        """
        default_queue = self.setup_default_queue()

        for c in self._consumers:
            if c.queue is None:
                queue = default_queue
            else:
                try:
                    queue = self._queues[c.queue]
                    if queue.exclusive or c.exclusive:
                        raise ValueError(
                            'Consumer %s is set exclusive with queue %s, but there '
                            'are other consumers already exist.' % (c.name,
                                                                    queue.name))
                except KeyError:
                    queue = self._setup_queue(c.queue)

            queue.add_consumer(c)

        # setup the queue on RabbitMQ
        for queue_name in self._queues.keys():
            self._channel.queue_declare(queue_name,  auto_delete=self.auto_delete, durable=self.durable)
            self._channel.queue_bind(queue_name, exchange=self._exchange)

        self.start_consuming()

    def start_consuming(self):
        if self._threaded:
            self._channel.add_on_cancel_callback(self.on_consumer_cancelled)
        else:
            pass
        for queue in self._queues.values():
            consumer_tag = self._channel.basic_consume(queue.name, queue.dispatcher)#, auto_ack=True)
            queue.dispatcher.consumer_tag = consumer_tag

        logger.info(self._queues)
        logger.info('Start consuming..')

        if not self._threaded:
            self._channel.start_consuming()

    def on_consumer_cancelled(self, method_frame):
        """Invoked by pika when RabbitMQ sends a Basic.Cancel for a consumer
        receiving messages.

        :param pika.frame.Method method_frame: The Basic.Cancel frame

        """
        logger.info('Consumer was cancelled remotely, shutting down: %r',
                    method_frame)
        self.close_channel()

    def run(self):
        """Run by connecting and then starting the IOLoop."""

        # make sure one processor one connection
        if self._threaded:
            with self._lock:
                self._connection = self.connect()
            self._connection.ioloop.start()
        else:
            self._connection = self.connect()
            self._channel = self._connection.channel()
            self._channel.exchange_declare(self._exchange, exchange_type='direct', auto_delete=self.auto_delete, durable=self.durable)
            self.on_exchange_declareok(None)

    def stop(self):
        """Cleanly shutdown the connection to RabbitMQ by stopping the consumer
        with RabbitMQ. When RabbitMQ confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when CTRL-C is pressed raising a KeyboardInterrupt exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with RabbitMQ. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.

        """
        self._closing = True
        self.close_channel()
        self.close_connection()



# -*- coding: utf-8 -*-
import os
from setuptools import find_packages, setup
import sys
import io

PY2 = sys.version_info[0] == 2

with io.open(os.path.join(os.path.dirname(__file__), 'requirements.txt'), encoding='utf-8') as f:
    requires = f.readlines()
    if PY2:
        requires += ['futures==3.2.0', 'monotonic==1.5']

with io.open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as f:
    README = f.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))


setup(
    name='rabbitmq-rpc',
    version='0.1.7',
    packages=find_packages(),
    include_package_data=True,
    description='A modified rabbit-rpc of https://github.com/MidTin/rabbit-rpc',
    long_description=README,
    long_description_content_type="text/markdown",
    author='leo',
    author_email='liupgd@foxmail.com',
    url='https://github.com/liupgd/rabbitmq_rpc',
    license='MIT',
    install_requires=requires,
    platforms = 'any',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: Implementation :: CPython',
    ],
    entry_points={
        'console_scripts': [
            'rabbitmq_rpc=rabbitmq_rpc:main',
        ]
    }
)



# -*- coding: utf-8 -*-
import logging
from threading import Lock
import sys
import pika
import warnings

logger = logging.getLogger(__name__)

class Connector(object):
    '''
    Base connector.
    Parameters:
    amqp_url: amqp url to rabbitmq server, format is like this: amqp://yourname:yourpasswd@serverip:serverport/
    host,port,username, passwd: If amqp_url is not specified, you should specify these server parameters.
    exchange: exchange in the server
    threaded: If true, connector will run with pika.SelectionConnection. It's a asynchronous io. Otherwise,
        pika.BlockConnection will be used. In asynchronous io mode. The RPC server can run in multi-threaded mode.
        There is no difference for RPC clients.
    kwargs: Invalid if amqp_url specified. See more from pika.ConnectionParameters
    '''
    DEFUALT_QUEUE = 'default'
    EXCHANGE_TYPE = 'direct'
    def __init__(self, amqp_url=None,
                 host = "localhost", port = 5672, prefetch_count = 1,
                 username = "guest", passwd="guest", exchange='default', threaded = True, auto_delete = True, durable = False,
                 **kwargs):
        self.auto_delete = auto_delete
        self.durable = durable
        self._url = amqp_url
        self._exchange = exchange
        self._host = host
        self._port = port
        self._username = username
        self._passwd = passwd
        self.prefetch_count = prefetch_count
        self._channel = None
        self._connection = None
        self._closing = False
        self._lock = Lock()
        if threaded and sys.platform == 'win32':
            warnings.warn("### In windows, pika may has problems with multi-thread processing ###")
        self._threaded = threaded
        if 'conn_parameters' in kwargs:
            self.conn_parameters = kwargs.get("conn_parameters")
        else:
            self.conn_parameters = None
        if amqp_url is not None:
            self.conn_parameters = pika.URLParameters(self._url)
        elif self.conn_parameters is None:
            self.conn_parameters = pika.ConnectionParameters(host= self._host, port=self._port,
                                                             credentials=pika.PlainCredentials(self._username, self._passwd),
                                                             **kwargs)



    def connect(self):
        """This method connects to RabbitMQ, returning the connection handle.
        When the connection is established, the on_connection_open method
        will be invoked by pika.

        :rtype: pika.SelectConnection

        """
        if self._threaded:
            self._connection = pika.SelectConnection(
                self.conn_parameters,
                on_open_callback=self.on_connection_open,
                on_close_callback=self.on_connection_closed)
        else:
            self._connection = pika.BlockingConnection(self.conn_parameters)
        return self._connection

    def on_connection_open(self, unused_connection):
        """This method is called by pika once the connection to RabbitMQ has
        been established. It passes the handle to the connection object in
        case we need it, but in this case, we'll just mark it unused.

        :type unused_connection: pika.SelectConnection

        """
        logger.info('Connection opened..')
        self.open_channel()

    def on_connection_closed(self, *args):#connection, reply_code, reply_text):
        """This method is invoked by pika when the connection to RabbitMQ is
        closed unexpectedly. Since it is unexpected, we will reconnect to
        RabbitMQ if it disconnects.

        :param pika.connection.Connection connection: The closed connection obj
        :param int reply_code: The server provided reply_code if given
        :param str reply_text: The server provided reply_text if given

        """
        self._channel = None
        if self._closing:
            self._connection.ioloop.stop()
        else:
            logger.info(
                'Connection was closed unexpected, we will try to reconnect it..')
            # self._connection.add_timeout(1, self.reconnect)

    def reconnect(self):
        """Will be invoked by the IOLoop timer if the connection is
        closed. See the on_connection_closed method.

        """
        # This is the old connection IOLoop instance, stop its ioloop
        self._connection.ioloop.stop()

        if not self._closing:

            # Create a new connection
            self._connection = self.connect()

            # There is now a new connection, needs a new ioloop to run
            self._connection.ioloop.start()

    def open_channel(self):
        self._connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """This method is invoked by pika when the channel has been opened.
        The channel object is passed in so we can make use of it.

        Since the channel is now open, we'll declare the exchange to use.

        :param pika.channel.Channel channel: The channel object

        """
        logger.info('Channel opened..')

        self._channel = channel
        self._channel.basic_qos(prefetch_count=self.prefetch_count)
        self.add_on_channel_close_callback()
        if self._exchange:
            self.setup_exchange(self._exchange)

    def add_on_channel_close_callback(self):
        """This method tells pika to call the on_channel_closed method if
        RabbitMQ unexpectedly closes the channel.

        """
        self._channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, chn_exception):#channel, reply_code, reply_text):
        """Invoked by pika when RabbitMQ unexpectedly closes the channel.
        Channels are usually closed if you attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, we'll close the connection
        to shutdown the object.

        :param pika.channel.Channel: The closed channel
        :param int reply_code: The numeric reason the channel was closed
        :param str reply_text: The text reason the channel was closed

        """
        logger.info('Channel closed..')
        self.close_connection()

    def setup_exchange(self, exchange_name):
        """Setup the exchange on RabbitMQ by invoking the Exchange.Declare RPC
        command. When it is complete, the on_exchange_declareok method will
        be invoked by pika.

        :param str|unicode exchange_name: The name of the exchange to declare

        """
        self._channel.exchange_declare(
            exchange_name, exchange_type=self.EXCHANGE_TYPE, auto_delete=self.auto_delete,
            callback=self.on_exchange_declareok, durable=self.durable)

    def on_exchange_declareok(self, unused_frame):
        """Invoked by pika when RabbitMQ has finished the Exchange.Declare RPC
        command.

        :param pika.Frame.Method unused_frame: Exchange.DeclareOk response frame

        """
        pass

    def close_connection(self):
        """This method closes the connection to RabbitMQ."""
        self._connection.close()

    def close_channel(self):
        """Invoke this command to close the channel with RabbitMQ by sending
        the Channel.Close RPC command.

        """
        if self._channel is not None:
            self._channel.close()





# -*- coding: utf-8 -*-
from argparse import ArgumentParser
import sys


class BaseCommand(object):

    help = ''

    def __init__(self):
        self.arg_parser = ArgumentParser(prog=self.__class__.__name__)
        self.add_arguments(self.arg_parser)

    def add_arguments(self, parser):
        pass

    def print_help(self):
        self.arg_parser.print_help()

    def run_from_argv(self, *args):
        options = vars(self.arg_parser.parse_args(args))

        try:
            self.execute(**options)
        except Exception as ex:
            sys.stderr.write('%s: %s' % (self.__class__.__name__, ex))
            sys.exit(1)

    def execute(self, **options):
        pass



# -*- coding: utf-8 -*-
from __future__ import absolute_import

import importlib
import imp
import logging
import os
import sys
import traceback

import pika

from rabbit_rpc.consumer import Consumer
from rabbit_rpc.credentials import AliyunCredentialsProvider
from rabbit_rpc.server import RPCServer
from .base import BaseCommand

logger = logging.getLogger(__name__)


class Worker(BaseCommand):

    name = 'worker'

    def add_arguments(self, parser):
        parser.add_argument(
            '-Q',
            '--queue',
            default='default',
            help='setup and bind to the specified queue')
        parser.add_argument(
            '--amqp',
            default='amqp://guest:guest@localhost:5672/',
            help='specify the broker url')
        parser.add_argument(
            '--django', help='setup django')
        parser.add_argument(
            '--alicert', help='''
            specify the certification of aliyun amqp in format:

            [access_key]:[access_secret]:[resource_owner_id]
            ''')

    def install_django(self, project_name):
        import django

        os.environ['DJANGO_SETTINGS_MODULE'] = project_name + '.settings'
        django.setup()

    def parse_aliyun_cert(self, raw_cert):
        try:
            access_key, access_secret, resource_owner_id = raw_cert.split(':')
        except ValueError:
            raise ValueError('invalid alicert')

        provider = AliyunCredentialsProvider(
            access_key, access_secret, resource_owner_id)

        return provider.get_username(), provider.get_password()

    def find_consumers(self, related_name='consumers'):
        path = os.getcwd()
        logger.info('Finding consumers...')

        consumers = []
        for dirpath in os.listdir(path):
            if dirpath.startswith('.'):
                continue

            if os.path.isdir(dirpath):
                try:
                    module = find_related_module(dirpath, related_name)
                    if module:
                        for item in dir(module):
                            c = getattr(module, item)
                            if isinstance(c, Consumer):
                                logger.info('[Consumer] %s.%s.%s', dirpath,
                                            related_name, c.name)
                                consumers.append(c)
                except ImportError:
                    pass

        return consumers

    def execute(self, **options):
        sys.path.append(os.getcwd())

        try:
            conn_parameters = pika.URLParameters(options['amqp'])

            if options.get('django'):
                self.install_django(options['django'])

            if options.get('alicert'):
                username, password = self.parse_aliyun_cert(options['alicert'])
                conn_parameters.credentials = pika.PlainCredentials(
                    username, password, erase_on_connect=True)

            consumers = self.find_consumers()
            if not consumers:
                sys.stderr.write('No consumer was detected.\n')
                sys.exit(1)

            server = RPCServer(
                consumers = consumers, conn_parameters=conn_parameters,
                queue_name=options['queue'])
            server.run()
        except KeyboardInterrupt:
            server.stop()
        except Exception:
            traceback.print_exc()
            sys.exit(1)


def find_related_module(package, related_name):
    """Find module in package."""
    try:
        importlib.import_module(package)
    except ImportError:
        package, _, _ = package.rpartition('.')
        if not package:
            raise

    try:
        pkg_path = importlib.import_module(package).__path__
    except AttributeError:
        return

    if not isinstance(pkg_path, list):
        pkg_path = pkg_path._path

    try:
        imp.find_module(related_name, pkg_path)
    except ImportError:
        return

    return importlib.import_module('{0}.{1}'.format(package, related_name))
